#! /usr/bin/python
# Copyright 2015,2016, Cumulus Networks Inc.  all rights reserved

try:
    import poe_base
    import time
except ImportError, e:
    raise ImportError (str(e) + 'required module not found')

#############################################################################
# POE chip specific base class, inherits from the base_poe class,
# provides acces to the following POE chips:
#   Broadcom 59103
#   Broadcom 59121
#

''' offsets within the BCM591xx communication packet
'''
BCM591xx_PKT_OFFSET_CMD=0
BCM591xx_PKT_OFFSET_SEQ=1
BCM591xx_PKT_OFFSET_DATA0=2
BCM591xx_PKT_OFFSET_DATA1=3
BCM591xx_PKT_OFFSET_DATA2=4
BCM591xx_PKT_OFFSET_DATA3=5
BCM591xx_PKT_OFFSET_DATA4=6
BCM591xx_PKT_OFFSET_DATA5=7
BCM591xx_PKT_OFFSET_DATA6=8
BCM591xx_PKT_OFFSET_DATA7=9
BCM591xx_PKT_OFFSET_DATA8=10
BCM591xx_PKT_OFFSET_CKSUM=11

BCM591xx_PKT_LEN=12

''' the two chip types supported
'''
BCM591xx_CHIP_59103=1
BCM591xx_CHIP_59121=2

''' command codes
'''
BCM591xx_CMD_SET_PORT_POE_ENABLE=0x00
BCM591xx_CMD_SET_PORT_ALLOW_POWER_UP=0x01
BCM591xx_CMD_SET_LOGICAL_MAP_ENABLE=0x02
BCM591xx_CMD_SET_PORT_RESET=0x03
BCM591xx_CMD_RESET_POE_SYSTEM=0x09
BCM591xx_CMD_SET_PORT_DETECTION_TYPE=0x10
BCM591xx_CMD_SET_PORT_PWR_THRESHOLD_TYPE=0x15
BCM591xx_CMD_SET_PORT_MAX_POWER=0x16
BCM591xx_CMD_SET_POWER_MODE=0x17
BCM591xx_CMD_SET_PWR_SOURCE_CFG=0x18
BCM591xx_CMD_SET_PORT_PRIORITY=0x1a
BCM591xx_CMD_SET_PORT_POWER_UP_MODE=0x1c
BCM591xx_CMD_SET_LOGICAL_PORT_MAP=0x1d
BCM591xx_CMD_GET_SYS_STATUS=0x20
BCM591xx_CMD_GET_PORT_STATUS=0x21
BCM591xx_CMD_GET_SYS_POWER=0x23
BCM591xx_CMD_GET_PORT_CONFIG=0x25
BCM591xx_CMD_GET_PORT_EXT_CONFIG=0x26
BCM591xx_CMD_GET_MULTIPORT_STATUS=0x28
BCM591xx_CMD_GET_PORT_GROUP_STATUS=0x2a        # BCM 59121 only
BCM591xx_CMD_GET_PORT_MEASUREMENTS=0x30
BCM591xx_CMD_SET_SYSCONFIGURATION=0xe0

BCM591xx_SUBCMD_SET_SYSCONFIGURATION=0xf0

''' command specific codes
'''
BCM591xx_POE_ENABLE=1
BCM591xx_POE_DISABLE=0

BCM591xx_POE_AF_MODE=0
BCM591xx_POE_AT_MODE=3

BCM591xx_POE_RESET=1

BCM591xx_DETECT_TYPE_NONE=0
BCM591xx_DETECT_TYPE_LEGACY_ONLY=1
BCM591xx_DETECT_TYPE_4P_ONLY=2
BCM591xx_DETECT_TYPE_4P_PLUS_LEGACY=3
BCM591xx_DETECT_TYPE_2P_ONLY=4
BCM591xx_DETECT_TYPE_2P_PLUS_LEGACY=5

BCM591xx_PWR_THRESHOLD_TYPE_NONE=0
BCM591xx_PWR_THRESHOLD_TYPE_CLASS_BASED=1
BCM591xx_PWR_THRESHOLD_TYPE_USER=2

BCM591xx_MODE_NONE=0
BCM591xx_MODE_STATIC_WITH_PRIORITY=1
BCM591xx_MODE_DYNAMIC_WITH_PRIORITY=2
BCM591xx_MODE_STATIC_NO_PRIORITY=3
BCM591xx_MODE_DYNAMIC_NO_PRIORITY=4

BCM591xx_PORT_PRIORITY_LOW=0
BCM591xx_PORT_PRIORITY_NORMAL=1
BCM591xx_PORT_PRIORITY_HIGH=2
BCM591xx_PORT_PRIORITY_CRITICAL=3

''' status and error codes
'''
BCM591xx_PORT_STATUS_DISABLED=0x00
BCM591xx_PORT_STATUS_SEARCHING=0x01
BCM591xx_PORT_STATUS_DELIVERING_POWER=0x02
BCM591xx_PORT_STATUS_TEST_MODE=0x03
BCM591xx_PORT_STATUS_FAULT=0x04
BCM591xx_PORT_STATUS_OTHER_FAULT=0x05
BCM591xx_PORT_STATUS_REQUESTING_POWER=0x06
BCM591xx_PORT_STATUS_DENIED=0x99

BCM591xx_PORT_ERROR_MPS_ABSENT=0x01
BCM591xx_PORT_ERROR_SHORT=0x02
BCM591xx_PORT_ERROR_OVERLOAD=0x03
BCM591xx_PORT_ERROR_POWER_DENIED=0x04
BCM591xx_PORT_ERROR_THERMAL_SHUTDOWN=0x05
BCM591xx_PORT_ERROR_STARTUP_FAILURE=0x06
BCM59121_PORT_ERROR_OVERVOLTAGE=0x00
BCM59121_PORT_ERROR_UNVERDOLTAGE=0x08

BCM591xx_DEVICE_TYPE_NONE=0x00
BCM591xx_DEVICE_TYPE_IEEEAF=0x01
BCM591xx_DEVICE_TYPE_LEGACY=0x02
BCM591xx_DEVICE_TYPE_INVALID=0x03

BCM591xx_PORT_RESET=1

bcm591xx_priority_map = {
    BCM591xx_PORT_PRIORITY_LOW:poe_base.PORT_CFG_PRIORITY_LOW,
    BCM591xx_PORT_PRIORITY_NORMAL:poe_base.PORT_CFG_PRIORITY_HIGH,
    BCM591xx_PORT_PRIORITY_HIGH:poe_base.PORT_CFG_PRIORITY_CRITICAL,
    BCM591xx_PORT_PRIORITY_CRITICAL:poe_base.PORT_CFG_PRIORITY_CRITICAL }

bcm591xx_port_error_map = {
    BCM591xx_CHIP_59103: {
        BCM591xx_PORT_ERROR_MPS_ABSENT:'MPS absent',
        BCM591xx_PORT_ERROR_SHORT:'short',
        BCM591xx_PORT_ERROR_OVERLOAD:'overload',
        BCM591xx_PORT_ERROR_POWER_DENIED:'power denied',
        BCM591xx_PORT_ERROR_THERMAL_SHUTDOWN:'thermal shutdown',
        BCM591xx_PORT_ERROR_STARTUP_FAILURE:'startup failure'
    },
    BCM591xx_CHIP_59121: {
        BCM59121_PORT_ERROR_UNVERDOLTAGE:'undervoltage',
        BCM591xx_PORT_ERROR_MPS_ABSENT:'MPS absent',
        BCM591xx_PORT_ERROR_SHORT:'short',
        BCM591xx_PORT_ERROR_OVERLOAD:'overload',
        BCM591xx_PORT_ERROR_POWER_DENIED:'power denied',
        BCM591xx_PORT_ERROR_THERMAL_SHUTDOWN:'thermal shutdown',
        BCM591xx_PORT_ERROR_STARTUP_FAILURE:'startup failure',
        BCM59121_PORT_ERROR_OVERVOLTAGE:'overvoltage'}
    }

class PoeBcm591xx(poe_base.PoeBase):

    class_based_power_map = {
        '0':16200, '1':4200, '2':7400, '3':16200, '4':31200
    }

    def __init__(self, chip):
        poe_base.PoeBase.__init__(self)
        self.chip = chip
        self.seq = 0

    ''' POE Interface Routines
    '''
    def get_poe_system_status(self, incremental=False):
        self.log_debug('bcm591xx::get_poe_system_status(%s)' % (str(incremental)))
        result = Bcm591xxGetSysStatusResult()
        self._handleSysRequest(BCM591xx_CMD_GET_SYS_POWER, result)

        port_results = {}
        if self.chip == BCM591xx_CHIP_59121:
            for group in range(6):
                request = self._new_request_str([BCM591xx_CMD_GET_PORT_GROUP_STATUS, 0x00, group])
                response = self._send_request(request)
                for port_idx in range(8):
                    port = group * 8 + port_idx
                    port_results[port] = Bcm591xxPortFullStatusResult(port, self.chip)
                    port_results[port].handle_response(response)
        else:
            for port in range(48):
                port_results[port] = Bcm591xxPortFullStatusResult(port, self.chip)
            self._handleQuadPortRequest(list(range(48)), BCM591xx_CMD_GET_MULTIPORT_STATUS, port_results)
        for port in range(48):
            result.ports[port] = port_results[port]._json()
        return result._json()

    def get_port_extra_status(self, portList):
        self.log_debug('poe_bm591xx::get_port_extra_status(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        results = {}
        for port in portList:
            result = Bcm591xxPortExtraStatusResult(port)
            self._handlePortCommand(port, BCM591xx_CMD_GET_PORT_MEASUREMENTS, result)
            self._handlePortCommand(port, BCM591xx_CMD_GET_PORT_EXT_CONFIG, result)
            results[port] = result._json()
        return results

    def get_poe_version_info(self):
        self.log_debug('poe_bcm591xx::get_poe_version()')
        return self._handleSysRequest(BCM591xx_CMD_GET_SYS_STATUS, Bcm591xxGetSwVersionResult())

    def set_port_priority(self, portList, level):
        self.log_debug('poe_bm591xx::set_port_priority((%s:%u)' % (str(portList), level))
        priority = bcm591xx_priority_map.get(level, -1)
        if priority == -1:
            raise poe_base.POEDRuntimeError('Attempt to set invalid priority level %u' % (level))
        portList = self._mk_list(portList)
        self._handleQuadPortCommand(portList, BCM591xx_CMD_SET_PORT_PRIORITY, level)

    def set_port_enable_poe(self, portList):
        self.log_debug('poe_bm591xx::setPortEnablePoe((%s)' % (str(portList)))
        portList = self._mk_list(portList)
        self._handleQuadPortCommand(portList, BCM591xx_CMD_SET_PORT_POE_ENABLE, BCM591xx_POE_ENABLE)

    def set_port_disable_poe(self, portList):
        self.log_debug('poe_bm591xx::setPortDisablePoe((%s)' % (str(portList)))
        portList = self._mk_list(portList)
        self._handleQuadPortCommand(portList, BCM591xx_CMD_SET_PORT_POE_ENABLE, BCM591xx_POE_DISABLE)

    def reset_poe_system(self):
        self.log_debug('poe_bcm591xx::reset_poe_system()')
        self._handleSysCommand(BCM591xx_CMD_RESET_POE_SYSTEM, [BCM591xx_POE_RESET])

    def reset_ports(self, portList):
        portList = self._mk_list(portList)
        reset_ports = []
        for port_id in portList:
            result = Bcm591xxPortFullStatusResult(port_id, self.chip)
            self._handlePortCommand(port_id, BCM591xx_CMD_GET_PORT_STATUS, result)
            if result.status == result.PORT_STATUS_DISABLED:
                continue
            reset_ports.append(port_id)
            resp = self._send_request(self._new_request_str([BCM591xx_CMD_SET_PORT_POE_ENABLE, 0,
                                                         port_id,
                                                         BCM591xx_POE_DISABLE]))
        time.sleep(5)
        for port_id in reset_ports:
            resp = self._send_request(self._new_request_str([BCM591xx_CMD_SET_PORT_POE_ENABLE, 0,
                                                         port_id,
                                                         BCM591xx_POE_ENABLE]))


    def save_poe_configuration(self):
        return self._handleSysSubCommand(BCM591xx_CMD_SET_SYSCONFIGURATION, BCM591xx_SUBCMD_SET_SYSCONFIGURATION)

    def get_poe_diag_info(self):
        self.log_debug('bcm591xx::get_poe_diag_info')
        result = Bcm591xxGetDiagInfoResult()
        for port_num in range(48):
            port_result = Bcm591xxGetDiagInfoPortResult(port_num)
            port_result.handle_response(self._send_request(self._new_request_str([BCM591xx_CMD_GET_PORT_CONFIG, 0, port_num])))
            port_result.handle_response(self._send_request(self._new_request_str([BCM591xx_CMD_GET_PORT_STATUS, 0, port_num])))
            result.ports[port_num] = port_result._json()
        self._handleSysRequest(BCM591xx_CMD_GET_SYS_POWER, result)
        return result._json()

    def set_port_l2_power(self, portList, allocated, requested):
        portList = self._mk_list(portList)
        # convert from milliwatts to 0.2 watts
        watts = int(requested) / 200
        self._handleQuadPortCommand(portList,
                                    BCM591xx_CMD_SET_PORT_PWR_THRESHOLD_TYPE,
                                    BCM591xx_PWR_THRESHOLD_TYPE_USER)
        self._handleQuadPortCommand(portList,
                                    BCM591xx_CMD_SET_PORT_MAX_POWER,
                                    watts)

    def exit_port_l2_mode(self, portList):
        self._handleQuadPortCommand(self._mk_list(portList),
                                    BCM591xx_CMD_SET_PORT_PWR_THRESHOLD_TYPE,
                                    BCM591xx_PWR_THRESHOLD_TYPE_CLASS_BASED)

    def get_port_4pair_state(self, portList):
        # only support 2 pair
        results = {}
        for port_id in self._mk_list(portList):
            results[port_id] = Bcm591xxGetPort4pairStateResult()._json()
        return results

    ''' poe_bm591xx utility routines
    '''
    def _handleSysCommand(self, cmd_id, vals=[]):
        self.log_debug('poe_bm591xx::_handleSysCommand(%u, %s)' % (cmd_id, str(vals)))
        req_list = [cmd_id, 0]
        for val in vals:
            req_list.append(val)
        self._send_request(self._new_request_str(req_list))

    def _handleSysRequest(self, cmd_id, result, vals=[]):
        self.log_debug('poe_bm591xx::_handleSysRequest(%u, %s)' % (cmd_id, str(vals)))
        req_list = [cmd_id, 0]
        for val in vals:
            req_list.append(val)
        resp = self._send_request(self._new_request_str(req_list))
        result.handle_response(resp)
        return result._json()

    def _handlePortCommand(self, port, cmd_id, result):
        self.log_debug('poe_bm591xx::_handlePortCommand(%u, 0x%02X)' % (port, cmd_id))
        req = self._new_request_str([cmd_id, 0, port])
        resp = self._send_request(req)
        result.handle_response(resp)
        return result._json()

    def _handleQuadPortCommand(self, portList, cmd_id, portVal):
        self.log_debug('poe_bm591xx::_handleQuadPortCommand(%s, 0x%02X, 0x%02X)' % (str(portList), cmd_id, portVal))
        request = self._new_request_str([cmd_id])
        idx = 2
        for port in portList:
            request[idx] = port
            request[idx+1] = portVal
            idx += 2
            if idx == 10:
                self._send_request(request)
                request = self._new_request_str([cmd_id])
                idx = 2
        if idx != 2:
            self._send_request(request)

    def _handleQuadPortRequest(self, portList, cmd_id, results):
        self.log_debug('poe_bm591xx::_handleQuadPortRequest(%s, 0x%02X)' % (str(portList), cmd_id))
        request = self._new_request_str([cmd_id])
        idx = 2
        result_idx = 0
        for port in portList:
            request[idx] = port
            request[idx+1] = 1
            idx += 2
            if idx == 10:
                response = self._send_request(request)
                for i in range(4):
                    results[result_idx].handle_response(response)
                    result_idx += 1
                idx = 2
                request = self._new_request_str([cmd_id])
        if idx != 2:
            for i in range(4):
                response = self._send_request(request)
                results[result_idx].handle_response(response)
                result_idx += 1

    def _handleSysSubCommand(self, cmd_id, sub_cmd_id):
        req_list = [cmd_id, sub_cmd_id]
        return self._send_request(self._new_request_str(req_list, True))

    def _handleMapPortsCommand(self, portMapList):
        logical_port = 0
        req_idx = 2
        request = self._new_request_str([BCM591xx_CMD_SET_LOGICAL_PORT_MAP])
        for port in portMapList:
            request[req_idx] = logical_port
            request[req_idx+1] = port
            req_idx += 2
            if req_idx == 10:
                result = self._send_request(request)
                request = self._new_request_str([BCM591xx_CMD_SET_LOGICAL_PORT_MAP])
                req_idx = 2
        if req_idx != 2:
            result = self._send_request(request)
        return result

    def _new_request_str(self, cmd_list, no_sequence=False):
        newRequest = []
        for i in range(len(cmd_list)):
            newRequest.append(cmd_list[i])
        for i in range(BCM591xx_PKT_OFFSET_CKSUM - len(newRequest)):
            newRequest.append(0xff)
        if not no_sequence:
            newRequest[BCM591xx_PKT_OFFSET_SEQ] = self.seq
            self.seq += 1
            if self.seq == 0xff:
                self.seq = 0
        return newRequest

    def _send_request(self, request):
        retries = 3
        request.append(self._calc_cksum(request))
        while retries:
            try:
                self.lock()
                self._send(request)
                response = self._get_response()
                self.unlock()
                self._verify_response(response, request)
                return response
            except poe_base.POEDRuntimeError, e:
                retries -= 1
        raise e

    def _send(self, request):
        self.log_verbose('tx > %s' % (self._bytes_to_str(request)))
        rlen = len(request)
        if rlen != BCM591xx_PKT_LEN:
            raise poe_base.POEDRuntimeError('Attempt to send invalid length request - expecting 12, got %u' % (rlen))
        self.poe_write(request)

    def _get_response(self):
        cstr = self.poe_read(BCM591xx_PKT_LEN)
        self.log_verbose('rx < %s' % (self._bytes_to_str(cstr)))
        return cstr

    def _clear_rx_queue(self):
        self.poe_read(BCM591xx_PKT_LEN)

    def _verify_response(self, response, request):
        if len(response) != BCM591xx_PKT_LEN:
            raise poe_base.POEDRuntimeError('invalid response length')
            self._clear_rx_queue()
        elif response[BCM591xx_PKT_OFFSET_SEQ] != request[BCM591xx_PKT_OFFSET_SEQ]:
            self._clear_rx_queue()
            raise poe_base.POEDSequenceError('request/response sequence numbers don\'t match [0x%02X:0x%02X]\n' %
                                     (response[BCM591xx_PKT_OFFSET_SEQ], request[BCM591xx_PKT_OFFSET_SEQ]))
        cksum = self._calc_cksum(response)
        if cksum != response[BCM591xx_PKT_OFFSET_CKSUM]:
            self._clear_rx_queue()
            raise poe_base.POEDRuntimeError('invalid checksum in response [0x%02X:0x%02X]' % (cksum, response[BCM591xx_PKT_OFFSET_CKSUM]))
        if response[BCM591xx_PKT_OFFSET_CMD] != request[BCM591xx_PKT_OFFSET_CMD]:
            self._clear_rx_queue()
            raise poe_base.POEDRuntimeError('invalid cmd ID in response')

    def _calc_cksum(self, request):
        self.log_debug('poe_bcm591xx::_calc_cksum()')
        rlen = min(len(request), BCM591xx_PKT_OFFSET_CKSUM)
        exp_len = BCM591xx_PKT_OFFSET_CKSUM
        if rlen != exp_len:
            raise poe_base.POEDRuntimeError("Calculating checksum on invalid length request - expecting %u, got %u" % (exp_len, rlen))

        sum = 0
        for i in range(rlen):
            sum += request[i]
        return sum & 0xff

class Bcm591xxPortFullStatusResult(poe_base.PortFullStatusResult):
        status_map = { BCM591xx_PORT_STATUS_DISABLED:poe_base.PortFullStatusResult.PORT_STATUS_DISABLED,
                       BCM591xx_PORT_STATUS_SEARCHING:poe_base.PortFullStatusResult.PORT_STATUS_SEARCHING,
                       BCM591xx_PORT_STATUS_DELIVERING_POWER:poe_base.PortFullStatusResult.PORT_STATUS_DELIVERING_POWER,
                       BCM591xx_PORT_STATUS_FAULT:poe_base.PortFullStatusResult.PORT_STATUS_FAULT,
                       BCM591xx_PORT_STATUS_OTHER_FAULT:poe_base.PortFullStatusResult.PORT_STATUS_FAULT,
                       BCM591xx_PORT_STATUS_REQUESTING_POWER:poe_base.PortFullStatusResult.PORT_STATUS_SEARCHING,
                       BCM591xx_PORT_STATUS_DENIED:poe_base.PortFullStatusResult.PORT_STATUS_DENIED}

        pd_type_map = { BCM591xx_DEVICE_TYPE_NONE:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_NONE,
                        BCM591xx_DEVICE_TYPE_IEEEAF:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_IEEE8023AF_AT,
                        BCM591xx_DEVICE_TYPE_LEGACY:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_LEGACY,
                        BCM591xx_DEVICE_TYPE_INVALID:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_INVALID }

        priority_map = {BCM591xx_PORT_PRIORITY_LOW:poe_base.PORT_CFG_PRIORITY_LOW,
                        BCM591xx_PORT_PRIORITY_NORMAL:poe_base.PORT_CFG_PRIORITY_HIGH,
                        BCM591xx_PORT_PRIORITY_HIGH:poe_base.PORT_CFG_PRIORITY_CRITICAL,
                        BCM591xx_PORT_PRIORITY_CRITICAL:poe_base.PORT_CFG_PRIORITY_CRITICAL}

        def __init__(self, port_num, chip):
            poe_base.PortFullStatusResult.__init__(self, port_num)
            self._chip = chip

        def handle_response(self, response):
            if response[BCM591xx_PKT_OFFSET_CMD] == BCM591xx_CMD_GET_PORT_STATUS:
                self.status = self.status_map.get(response[BCM591xx_PKT_OFFSET_DATA1], response[BCM591xx_PKT_OFFSET_DATA1])
                if self.status == poe_base.PortFullStatusResult.PORT_STATUS_FAULT:
                    self.error_str = bcm591xx_port_error_map.get(self._chip).get(response[BCM591xx_PKT_OFFSET_DATA2], 'other [0x%02X]' % (response[BCM591xx_PKT_OFFSET_DATA2]))
                elif self.status == poe_base.PortFullStatusResult.PORT_STATUS_DELIVERING_POWER:
                    self.pd_class = response[BCM591xx_PKT_OFFSET_DATA3]
                    self.pd_type = self.pd_type_map.get(response[BCM591xx_PKT_OFFSET_DATA4], response[BCM591xx_PKT_OFFSET_DATA4])
            elif response[BCM591xx_PKT_OFFSET_CMD] == BCM591xx_CMD_GET_MULTIPORT_STATUS or \
                            response[BCM591xx_PKT_OFFSET_CMD] == BCM591xx_CMD_GET_PORT_GROUP_STATUS:
                if response[BCM591xx_PKT_OFFSET_CMD] == BCM591xx_CMD_GET_MULTIPORT_STATUS:
                    data = response[(self.poe_port % 4) * 2 + 3]
                else:
                    data = response[(self.poe_port % 8) + 3]
                class_or_fault = (data & 0x70) >> 4
                status = data & 0x0f
                if status == BCM591xx_PORT_STATUS_REQUESTING_POWER:
                    # the bcm95121 doesn't ever seem to report power_denied just sits in
                    # requesting_power state forever
                    status = BCM591xx_PORT_STATUS_DENIED
                    class_or_fault = BCM591xx_PORT_ERROR_POWER_DENIED
                self.status = self.status_map.get(status, status)
                if self.status == poe_base.PortFullStatusResult.PORT_STATUS_FAULT or self.status == poe_base.PortFullStatusResult.PORT_STATUS_DENIED:
                    self.error_str = bcm591xx_port_error_map.get(self._chip).get(class_or_fault, 'other [0x%02X]' % (class_or_fault))
                elif self.status == poe_base.PortFullStatusResult.PORT_STATUS_DELIVERING_POWER:
                    self.pd_class = class_or_fault
                    if data & 0x80:
                        self.pd_type = self.pd_type_map[BCM591xx_DEVICE_TYPE_IEEEAF]
                    else:
                        self.pd_type = self.pd_type_map[BCM591xx_DEVICE_TYPE_LEGACY]
            elif response[BCM591xx_PKT_OFFSET_CMD] == BCM591xx_CMD_GET_PORT_EXT_CONFIG:
                self.priority = self.priority_map.get(response[BCM591xx_PKT_OFFSET_DATA4])

class Bcm591xxPortExtraStatusResult(poe_base.PoePortResult):
    priority_map = {BCM591xx_PORT_PRIORITY_LOW:poe_base.PORT_CFG_PRIORITY_LOW,
                    BCM591xx_PORT_PRIORITY_NORMAL:poe_base.PORT_CFG_PRIORITY_HIGH,
                    BCM591xx_PORT_PRIORITY_HIGH:poe_base.PORT_CFG_PRIORITY_CRITICAL,
                    BCM591xx_PORT_PRIORITY_CRITICAL:poe_base.PORT_CFG_PRIORITY_CRITICAL}
    def __init__(self, port):
        poe_base.PoePortResult.__init__(self, port)
        self.add_enum_maps({'priority':poe_base.PortFullStatusResult.cfg_priority_map })
        self.priority = None
        self.voltage = None
        self.current = None
        self.power = None

    def handle_response(self, response):
        if response[BCM591xx_PKT_OFFSET_CMD] == BCM591xx_CMD_GET_PORT_MEASUREMENTS:
            self.voltage = self._mk_word_data(response[BCM591xx_PKT_OFFSET_DATA1], response[BCM591xx_PKT_OFFSET_DATA2])
            self.voltage *= round(64.45/1000.0,1)                                                                                 # Volts
            self.current = self._mk_word_data(response[BCM591xx_PKT_OFFSET_DATA3], response[BCM591xx_PKT_OFFSET_DATA4])           # mA
            self.power = self._mk_word_data(response[BCM591xx_PKT_OFFSET_DATA7], response[BCM591xx_PKT_OFFSET_DATA8])/10.0        # Watts
        elif response[BCM591xx_PKT_OFFSET_CMD] == BCM591xx_CMD_GET_PORT_EXT_CONFIG:
            self.priority = self.priority_map.get(response[BCM591xx_PKT_OFFSET_DATA4])

class Bcm591xxGetSwVersionResult(poe_base.SwVersionResult):

    def handle_response(self, response):
        ver_byte = response[BCM591xx_PKT_OFFSET_DATA5]
        self.major_version = ver_byte >> 4
        self.minor_version = ver_byte & 0xf
        self.ver_str = '%u.%02u' % (self.major_version, self.minor_version)

class Bcm591xxGetSysStatusResult(poe_base.PoeSysStatusResult):

    def handle_response(self, response):
        self.sys_consumed_power = self._mk_word_data(response[BCM591xx_PKT_OFFSET_DATA0], response[BCM591xx_PKT_OFFSET_DATA1])/10.0    # Watts
        self.sys_power_limit = self._mk_word_data(response[BCM591xx_PKT_OFFSET_DATA2], response[BCM591xx_PKT_OFFSET_DATA3])/10.0    # Watts

class Bcm591xxGetDiagInfoResult(poe_base.PoeResult):
    def __init__(self):
        poe_base.PoeResult.__init__(self)
        self.ports = {}

    def handle_response(self, response):
        cmd = response[BCM591xx_PKT_OFFSET_CMD]
        if cmd == BCM591xx_CMD_GET_SYS_POWER:
            self.pwr_allocated = self._mk_word_data(response[BCM591xx_PKT_OFFSET_DATA0], response[BCM591xx_PKT_OFFSET_DATA1])/10.0
            self.pwr_available = self._mk_word_data(response[BCM591xx_PKT_OFFSET_DATA2], response[BCM591xx_PKT_OFFSET_DATA3])/10.0
            self.pwr_mpsm_status = response[BCM591xx_PKT_OFFSET_DATA4]
            self.pwr_ii_cfg = response[BCM591xx_PKT_OFFSET_DATA5]

class Bcm591xxGetDiagInfoPortResult(poe_base.PoePortResult):
    def __init__(self, port):
        poe_base.PoePortResult.__init__(self, port)
        self.add_enum_maps({'port_status':{BCM591xx_PORT_STATUS_DISABLED:'disabled', BCM591xx_PORT_STATUS_SEARCHING:'searching',
                                      BCM591xx_PORT_STATUS_DELIVERING_POWER:'delivering power', BCM591xx_PORT_STATUS_TEST_MODE:'test mode',
                                      BCM591xx_PORT_STATUS_FAULT:'fault', BCM591xx_PORT_STATUS_OTHER_FAULT:'other fault',
                                      BCM591xx_PORT_STATUS_REQUESTING_POWER:'requesting power'}})

    def handle_response(self, response):
        cmd = response[BCM591xx_PKT_OFFSET_CMD]
        if cmd == BCM591xx_CMD_GET_PORT_STATUS:
            self.port_status = response[BCM591xx_PKT_OFFSET_DATA1]
            self.port_err_type = response[BCM591xx_PKT_OFFSET_DATA2]
            self.port_class = response[BCM591xx_PKT_OFFSET_DATA3]
            self.port_pd_type = response[BCM591xx_PKT_OFFSET_DATA4]
            self.port_mps_status = '0x%02X' % response[BCM591xx_PKT_OFFSET_DATA5]
            self.port_pwr_mode = response[BCM591xx_PKT_OFFSET_DATA6]
        elif cmd == BCM591xx_CMD_GET_PORT_CONFIG:
            self.cfg_pse_enable = response[BCM591xx_PKT_OFFSET_DATA1]
            self.cfg_auto_mode = response[BCM591xx_PKT_OFFSET_DATA2]
            self.cfg_detection_type = response[BCM591xx_PKT_OFFSET_DATA3]
            self.cfg_classif_type = response[BCM591xx_PKT_OFFSET_DATA4]
            self.cfg_disconnect_type = response[BCM591xx_PKT_OFFSET_DATA5]
            self.cfg_pair_cfg = response[BCM591xx_PKT_OFFSET_DATA6]

class Bcm591xxGetPort4pairStateResult(poe_base.PoePort4portStateResult):
    def __init__(self):
        poe_base.PoePort4portStateResult.__init__(self)
        self.pair_mode = self.RESULT_2_PAIR
