#! /usr/bin/python
# Copyright 2015,2016, Cumulus Networks Inc.  all rights reserved

try:
    import time
    import poe_base
    import sys
except ImportError, e:
    raise ImportError (str(e) + 'required module not found')

#############################################################################
# POE chip specific base class, inherits from the base_poe class,
# provides access to the following POE chips:
#   MicroSemi PD69100
#   MicroSemi PD69200
#

''' offsets within the PD69100 communication packet
'''
PD69100_PKT_OFFSET_KEY=0
PD69100_PKT_OFFSET_SEQ=1
PD69100_PKT_OFFSET_TXSUB0=2
PD69100_PKT_OFFSET_TXSUB1=3
PD69100_PKT_OFFSET_TXSUB2=4
PD69100_PKT_OFFSET_TXDATA0=5
PD69100_PKT_OFFSET_TXDATA1=6
PD69100_PKT_OFFSET_TXDATA2=7
PD69100_PKT_OFFSET_TXDATA3=8
PD69100_PKT_OFFSET_TXDATA4=9
PD69100_PKT_OFFSET_TXDATA5=10
PD69100_PKT_OFFSET_TXDATA6=11
PD69100_PKT_OFFSET_TXDATA7=12
PD69100_PKT_OFFSET_CKSUM0=13
PD69100_PKT_OFFSET_CKSUM1=14

PD69100_PKT_OFFSET_RXDATA0=2
PD69100_PKT_OFFSET_RXDATA1=3
PD69100_PKT_OFFSET_RXDATA2=4
PD69100_PKT_OFFSET_RXDATA3=5
PD69100_PKT_OFFSET_RXDATA4=6
PD69100_PKT_OFFSET_RXDATA5=7
PD69100_PKT_OFFSET_RXDATA6=8
PD69100_PKT_OFFSET_RXDATA7=9
PD69100_PKT_OFFSET_RXDATA8=10
PD69100_PKT_OFFSET_RXDATA9=11
PD69100_PKT_OFFSET_RXDATA10=12

PD69100_PKT_LEN=15
PD69100_PKT_CKSUM_LEN=2

''' packet field codes
'''
PD69100_PKT_KEY_CMD=0x00
PD69100_PKT_KEY_PROG=0x01
PD69100_PKT_KEY_REQUEST=0x02
PD69100_PKT_KEY_TELEMETRY=0x03
PD69100_PKT_KEY_WRITE=0x04
PD69100_PKT_KEY_REPORT=0x52

PD69100_PKT_SUB_TEST=0x04
PD69100_PKT_SUB_CHANNEL=0x05
PD69100_PKT_SUB_E2=0x06
PD69100_PKT_SUB_GLOBAL=0x07
PD69100_PKT_SUB_FLASH=0xff
PD69100_PKT_SUB_RESTORE_DEFAULTS=0x2d

PD69100_PKT_SUB1_EN_DIS_4PAIR=0x02
PD69100_PKT_SUB1_PRIORITY=0x0a
PD69100_PKT_SUB1_SUPPLY=0x0b
PD69100_PKT_SUB1_PORTFULLINIT=0x0c
PD69100_PKT_SUB1_PORTSTATUS=0x0e
PD69100_PKT_SUB1_SAVE_CONFIG=0x0f
PD69100_PKT_SUB1_VERSIONZ=0x1e
PD69100_PKT_SUB1_MEASUREMENTS=0x25
PD69100_PKT_SUB1_PORTSTATUS1=0x31
PD69100_PKT_SUB1_PORTSTATUS2=0x32
PD69100_PKT_SUB1_PORTSTATUS3=0x33
PD69100_PKT_SUB1_SYSTEM_STATUS=0x3d
PD69100_PKT_SUB1_TEMP_MATRIX=0x43
PD69100_PKT_SUB1_ACTIVE_MATRIX=0x44
PD69100_PKT_SUB1_PORTSTATUS4=0x47
PD69100_PKT_SUB1_PORTSTATUS5=0x48
PD69100_PKT_SUB1_READ_RONS=0x52
PD69100_PKT_SUB1_WRITE_RONS=0x53
PD69100_PKT_SUB1_RESET=0x55
PD69100_PKT_SUB1_INDIVIDUAL_MASK=0x56
PD69100_PKT_SUB1_DEVICE_PARAMS=0x87
PD69100_PKT_SUB1_SET_LAYER2_PD=0xa6
PD69100_PKT_SUB1_GET_LAYER2_PD=0xa8
PD69100_PKT_SUB1_ALTB_POWER_UP=0xa9

PD69100_PKT_SUB2_MAIN=0x17
PD69100_PKT_SUB2_SW_VERSION=0x21
PD69100_PKT_SUB2_POWER_BUDGET=0x57
PD69100_PKT_SUB2_POWER_MGMT_MODE=0x5f
PD69100_PKT_SUB2_POWER_INFO=0x60

PD69100_PKT_SUB2_4PAIR=0x34

''' packet data definitions
'''
PD69100_PKT_DATA_NOP=0x4e
PD69100_PKT_DATA_CMDMASKBOTH=0xf0
PD69100_PKT_DATA_CMDDISABLE=0x00
PD69100_PKT_DATA_CMDENABLE=0x01

PD69100_PKT_DATA_AFMASK_AF=0x00
PD69100_PKT_DATA_AFMASK_AT=0x01
PD69100_PKT_DATA_AFMASK_POH=0x02

PD69100_PKT_DATA_PRIORITY_LOW=3
PD69100_PKT_DATA_PRIORITY_HIGH=2
PD69100_PKT_DATA_PRIORITY_CRITICAL=1

PD69100_PKT_DATA_PORT_DISABLED=0
PD69100_PKT_DATA_PORT_ENABLED=1

''' status codes for 'getAllPortsStatus' request
'''
PD69100_PORT_STATUS_OK_LEGACY=0x00
PD69100_PORT_STATUS_OK_AF=0x01
PD69100_PORT_STATUS_OK_AT=0x02
PD69100_PORT_STATUS_OK_2PAIR=0x03
PD69100_PORT_STATUS_OK_4PAIR=0x04
PD69100_PORT_STATUS_HW_DISABLED=0x08
PD69100_PORT_STATUS_DISABLED=0x1a
PD69100_PORT_DETECTION_IN_PROGRESS=0x1b
PD69100_PORT_INVALID_DEVICE=0x1c
PD69100_PORT_OVERUNDERLOAD=0x1d
PD69100_PORT_UNDERLOAD=0x1e
PD69100_PORT_OVERLOAD=0x1f
PD69100_PORT_POWER_BUDGET_EXCEEDED=0x20
PD69100_PORT_INTERNAL_FAULT=0x21
PD69100_PORT_BAD_CAP_DETECTED=0x25
PD69100_PORT_OVERVOLTAGE=0x2d
PD69100_PORT_UNDERVOLTAGE=0x2e
PD69100_PORT_SHORT=0x34
PD69100_PORT_OVERTEMPERATURE=0x35
PD69100_PORT_TOOHOT=0x36
PD69100_PORT_POWER_BUDGET_EXCEEDED_STATIC=0x3c
PD69100_PORT_STATUS_OK_HPD=0x40
PD69100_PORT_OVERPOWER=0x41
PD69100_PORT_INVALIDCLASS=0x43

''' individual mask codes
'''
PD69100_IND_MASK_IGN_HIGH_PRIORITY = 0x00
PD69100_IND_MASK_LEGACY_ENABLED=0x01
PD69100_IND_MASK_LEGACY_CLASS_MISMATCH=0x08
PD69100_IND_MASK_PSE_POWERING_PSE=0x1F
PD69100_IND_MASK_LAYER2_LLDP=0x2e
PD69100_IND_MASK_4PAIR_ENABLE=0x34
PD69100_IND_MASK_CLASS0_EQUAL_AF=0x38
PD69100_IND_MASK_CLASS_ERROR_EQUAL_0=0x3b

''' power management mode definitions
'''
PD69100_PM_MODE_PM1_DYNAMIC=0
PD69100_PM_MODE_PM1_1_3_STATIC=1
PD69100_PM_MODE_PM1_0_3_STATIC=2
PD69100_PM_MODE_PM1_1_4_STATIC=3
PD69100_PM_MODE_PM1_0_4_STATIC=4
PD69100_PM_MODE_PM1_LLDP=5
PD69100_PM_MODE_PM1_4_STATIC=6

PD69100_PM_MODE_PM2_PPL=0
PD69100_PM_MODE_PM2_CLASS=1
PD69100_PM_MODE_PM2_ICUT=2

PD69100_PM_MODE_PM3_NONE=0
PD69100_PM_MODE_PM3_CLASS_1_3=1
PD69100_PM_MODE_PM3_CLASS_0_3=2
PD69100_PM_MODE_PM3_CLASS_1_4=3
PD69100_PM_MODE_PM3_CLASS_0_4=4

''' altb power up definitions
'''
PD69100_ALTB_PUP_CFG_DISABLE=0
PD69100_ALTB_PUP_CFG_ENABLE=1
PD69100_ALTB_PUP_CMD_DISABLE=0
PD69100_ALTB_PUP_CMD_ENABLE=1

''' Firmware upgrade request/responses
'''
PD69100_BOOT_SECTION_RESPONSE="TPE"
PD69100_ERASE_MEMORY_RESPONSE="TOE"
PD69100_WAIT_FOR_ERASURE_SUCCESS_RESPONSE="TE"
PD69100_WAIT_FOR_ERASURE_SUCCESS_RESPONSE2="TPE"
PD69100_PROGRAM_MEMORY_RESPONSE="TOP"
PD69100_S19_LINE_RESPONSE="T*"
PD69100_S19_EOF_RESPONSE="TP"
PD69100_RST_REQUEST="RST"

pd69100_priority_map = { PD69100_PKT_DATA_PRIORITY_LOW:poe_base.PORT_CFG_PRIORITY_LOW,
                         PD69100_PKT_DATA_PRIORITY_HIGH:poe_base.PORT_CFG_PRIORITY_HIGH,
                         PD69100_PKT_DATA_PRIORITY_CRITICAL:poe_base.PORT_CFG_PRIORITY_CRITICAL}

pd69100_port_error_map = {
    PD69100_PORT_INVALID_DEVICE:'invalid device',
    PD69100_PORT_OVERUNDERLOAD:'overload/underload',
    PD69100_PORT_UNDERLOAD:'underload',
    PD69100_PORT_OVERLOAD:'overload',
    PD69100_PORT_POWER_BUDGET_EXCEEDED:'power budget exceeded',
    PD69100_PORT_INTERNAL_FAULT:'internal fault',
    PD69100_PORT_OVERVOLTAGE:'over voltage',
    PD69100_PORT_UNDERVOLTAGE:'under voltage',
    PD69100_PORT_SHORT:'short',
    PD69100_PORT_OVERTEMPERATURE:'over temperature',
    PD69100_PORT_TOOHOT:'too hot',
    PD69100_PORT_OVERPOWER:'over power',
    PD69100_PORT_INVALIDCLASS:'invalid class',
    PD69100_PORT_BAD_CAP_DETECTED:'detection failed',
    PD69100_PORT_POWER_BUDGET_EXCEEDED_STATIC:'power budget exceeded',
}

pd69100_port_ok_results = {
    PD69100_PORT_STATUS_OK_LEGACY:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69100_PORT_STATUS_OK_AF:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69100_PORT_STATUS_OK_AT:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69100_PORT_STATUS_OK_2PAIR:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69100_PORT_STATUS_OK_4PAIR:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69100_PORT_STATUS_OK_HPD:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69100_PORT_STATUS_HW_DISABLED:poe_base.PortStatusResult.PORT_STATUS_DISABLED,
    PD69100_PORT_STATUS_DISABLED:poe_base.PortStatusResult.PORT_STATUS_DISABLED,
    PD69100_PORT_DETECTION_IN_PROGRESS:poe_base.PortStatusResult.PORT_STATUS_SEARCHING,
    PD69100_PORT_POWER_BUDGET_EXCEEDED:poe_base.PortStatusResult.PORT_STATUS_DENIED,
    PD69100_PORT_POWER_BUDGET_EXCEEDED_STATIC:poe_base.PortStatusResult.PORT_STATUS_DENIED,
}

pd69100_port_pd_type_map = {
    PD69100_PORT_STATUS_OK_LEGACY:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_LEGACY,
    PD69100_PORT_STATUS_OK_AF:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_IEEE8023AF,
    PD69100_PORT_STATUS_OK_AT:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_IEEE8023AT,
    PD69100_PORT_STATUS_OK_2PAIR:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_2PAIR_LLDP,
    PD69100_PORT_STATUS_OK_4PAIR:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_4PAIR_LLDP,
    PD69100_PORT_STATUS_OK_HPD:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_HIGH_POWER,
}

# port events
PD69100_INT_PORT_EVENT_MASK = 0x3f

# system/device events
PD69100_INT_DEVICE_OVERTEMP1 =  1 << 8
PD69100_INT_DEVICE_OVERTEMP2 =  1 << 9
PD69100_INT_INTERNAL_DEVICE_FAULT = 1 << 10
PD69100_INT_MAIN_VOLTAGE_ERROR = 1 << 13
PD69100_INT_SYS_EVENT_MASK = PD69100_INT_DEVICE_OVERTEMP1 | PD69100_INT_DEVICE_OVERTEMP2 |\
    PD69100_INT_INTERNAL_DEVICE_FAULT | PD69100_INT_MAIN_VOLTAGE_ERROR


class PoePD69100(poe_base.PoeBase):
    class_based_power_map = {
    }

    def __init__(self):
        poe_base.PoeBase.__init__(self)
        self.seq = 0x10

    def poe_system_init(self):
        pass

    def get_poe_system_status(self, incremental=False):
        self.log_debug('PoePD69100::get_poe_system_status(%s)' % (str(incremental)))
        result = PD69100GetSysStatusResult()
        request = self._new_request_str([PD69100_PKT_KEY_REQUEST,
                                         0x00,
                                         PD69100_PKT_SUB_GLOBAL,
                                         PD69100_PKT_SUB1_SYSTEM_STATUS])
        response = self._send_request(request)
        result.handle_response(response, PD69100GetSysStatusResult.INTERRUPTS_RESPONSE)

        cmds_and_ports = [ (PD69100_PKT_SUB1_PORTSTATUS1, list(range(11))),
                           (PD69100_PKT_SUB1_PORTSTATUS2, list(range(11, 22))),
                           (PD69100_PKT_SUB1_PORTSTATUS3, list(range(22, 26))),
                           (PD69100_PKT_SUB1_PORTSTATUS4, list(range(26, 37))),
                           (PD69100_PKT_SUB1_PORTSTATUS5, list(range(37, 48)))]
        port_results = {}
        for port_cmd, ports in cmds_and_ports:
            request = self._new_request_str([PD69100_PKT_KEY_REQUEST,
                                            0x00,
                                            PD69100_PKT_SUB_GLOBAL,
                                            port_cmd])
            response = self._send_request(request)
            for port in ports:
                port_result = PD69100PortStatusResult(port)
                port_result.handle_response(response)
                port_results[port] = port_result._json()
        result.ports = port_results

        request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                       PD69100_PKT_SUB_GLOBAL,
                                       PD69100_PKT_SUB1_SUPPLY, PD69100_PKT_SUB2_MAIN])
        response = self._send_request(request)
        result.handle_response(response, PD69100GetSysStatusResult.SUPPLY_RESPONSE)

        for device_id in range(8):
            request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                           PD69100_PKT_SUB_GLOBAL,
                                           PD69100_PKT_SUB1_DEVICE_PARAMS, device_id])
            response = self._send_request(request)
            result.handle_response(response, PD69100GetSysStatusResult.TEMP_RESPONSE)

        return result._json()

    def get_port_extra_status(self, portList):
        self.log_debug('PoePD69100::get_port_extra_status(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        results = {}
        for port in portList:
            result = self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PORTSTATUS, port],
                                             PD69100PortFullStatusResult(port),
                                             PD69100PortFullStatusResult.PORT_STATUS_RESPONSE)
            self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_MEASUREMENTS, port],
                                             result, PD69100PortFullStatusResult.MEASUREMENTS_RESPONSE)
            self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PRIORITY, port],
                                             result, PD69100PortFullStatusResult.PRIORITY_RESPONSE)
            # the PD69200 sometimes reports non-zero values for current, power, voltage on non-connected ports
            if result.status != poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER:
                result.current = 0
                result.power = 0
                result.voltage = 0

            results[port] = result._json()
        return results

    def get_ports_4pair_config(self):
        self.log_debug('PoePD69100::get_port_4pair_config()')
        result = {}
        res = self._send_request(self._new_request_str([PD69100_PKT_KEY_REQUEST,
                                                   0x00,
                                                   PD69100_PKT_SUB_GLOBAL,
                                                   PD69100_PKT_SUB1_EN_DIS_4PAIR]))
        port = 0
        for byte in range(2, 9):
            if byte == 5:
                continue
            for bit in range(8):
                result[port] = '4pair' if res[byte] & (1 << bit) else '2pair'
                port += 1
        return result

    def get_port_4pair_state(self, portList):
        self.log_debug('PoePD69100::get_ports_4pair_state()')
        portList = self._mk_list(portList)

        results = {}
        for port in portList:
            res = self._handle_port_command([PD69100_PKT_KEY_REQUEST,
                                                       0x00,
                                                       PD69100_PKT_SUB_CHANNEL,
                                                       PD69100_PKT_SUB1_ALTB_POWER_UP,
                                                       port], PD69100GetPort4pairStateResult(port))
            results[port] = res._json()
        return results

    def get_port_priority(self, portList):
        self.log_debug('PoePD69100::get_port_priority(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        results = {}
        for port in portList:
            results[port] = self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PRIORITY, port],
                                             PD69100PortPriorityResult(port))._json()
        return results

    def get_poe_diag_info(self):
        self.log_debug('PoePD69100::get_poe_diag_info')
        result = PD69100GetDiagInfoResult()

        ''' get power info
        '''
        request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                       PD69100_PKT_SUB_GLOBAL,
                                       PD69100_PKT_SUB1_SUPPLY, PD69100_PKT_SUB2_MAIN])
        response = self._send_request(request)
        result.handle_response(response, PD69100GetDiagInfoResult.SUPPLY_RESPONSE)

        ''' get power management info
        '''
        request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                         PD69100_PKT_SUB_GLOBAL,
                                         PD69100_PKT_SUB1_SUPPLY,
                                         PD69100_PKT_SUB2_POWER_MGMT_MODE])
        response = self._send_request(request)
        result.handle_response(response, PD69100GetDiagInfoResult.MODE_RESPONSE)

        ''' get individual mask info
        '''
        masks = (0x00, 0x01, 0x08, 0x0c, 0x11, 0x14, 0x15, 0x1b, 0x1e, 0x1f, 0x28, 0x29, 0x2a, 0x2b, 0x2d, 0x2e, 0x2f,
                 0x30, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x43 )
        for mask in masks:
            request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                             PD69100_PKT_SUB_GLOBAL,
                                             PD69100_PKT_SUB1_INDIVIDUAL_MASK,
                                             mask])
            response = self._send_request(request)
            result.handle_response(response, PD69100GetDiagInfoResult.MASK_RESPONSE, mask)

        ''' get power-up timeout setting for all devices
        '''
        for device in range(0x0c):
            request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                             PD69100_PKT_SUB_TEST,
                                             PD69100_PKT_SUB1_READ_RONS,
                                             0, 0x99, device])
            response = self._send_request(request)
            result.handle_response(response, PD69100GetDiagInfoResult.REG_RESPONSE, device)

        ''' get port info
        '''
        for port in range(48):
            port_result = self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PORTSTATUS, port],
                                             PD69100GetPortDiagInfoResult(port),
                                             PD69100GetPortDiagInfoResult.STATUS_RESPONSE)

            self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_SUPPLY, port],
                                             port_result, PD69100GetPortDiagInfoResult.POWER_LIMIT_RESPONSE)
            self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_GET_LAYER2_PD, port],
                                             port_result, PD69100GetPortDiagInfoResult.L2_POWER_RESPONSE)
            result.ports[port] = port_result._json()
        return result._json()

    def get_poe_version_info(self):
        self.log_debug('PoePD69100::get_poe_version_info')
        result = self._handle_sys_command([PD69100_PKT_KEY_REQUEST, 0x00,
                                           PD69100_PKT_SUB_GLOBAL,
                                           PD69100_PKT_SUB1_VERSIONZ,
                                           PD69100_PKT_SUB2_SW_VERSION],
                                           PD69100GetSwVersionResult())
        return result._json()

    def set_port_priority(self, portList, level):
        self.log_debug('PoePD69100::set_port_priority(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        priority = PD69100_PKT_DATA_PRIORITY_LOW
        for our_val, base_val in pd69100_priority_map.items():
            if base_val == level:
                priority = our_val
        for port in portList:
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PRIORITY, port,
                                             priority])
            self._send_request(request)

    def set_port_enable_poe(self, portList):
        self.log_debug('PoePD69100::set_port_enable_poe(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        for port in portList:
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PORTFULLINIT, port,
                                             PD69100_PKT_DATA_CMDMASKBOTH | PD69100_PKT_DATA_CMDENABLE,
                                             PD69100_PKT_DATA_AFMASK_AT])
            self._send_request(request)

    def set_port_disable_poe(self, portList):
        self.log_debug('PoePD69100::set_port_disable_poe(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        for port in portList:
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PORTFULLINIT, port,
                                             PD69100_PKT_DATA_CMDMASKBOTH | PD69100_PKT_DATA_CMDDISABLE,
                                             PD69100_PKT_DATA_AFMASK_AT])
            self._send_request(request)

    def set_port_temp_matrix(self, port, phys_a, phys_b=0xff):
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_TEMP_MATRIX,
                                             port,
                                             phys_a, phys_b])
            self._send_request(request)

    def set_active_matrix(self):
        request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                         PD69100_PKT_SUB_GLOBAL,
                                         PD69100_PKT_SUB1_TEMP_MATRIX])
        self._send_request(request)

    def set_enable_4pair_mode(self, portList):
        portList = self._mk_list(portList)
        for port in portList:
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_EN_DIS_4PAIR,
                                             port,
                                             PD69100_PKT_DATA_CMDENABLE])
            self._send_request(request)

    def set_enable_4pair_power(self, portList):
        self.log_debug('PoePD69100::set_enable_4pair_power(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        for port in portList:
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_ALTB_POWER_UP,
                                             port,
                                             PD69100_ALTB_PUP_CFG_ENABLE,
                                             PD69100_ALTB_PUP_CMD_ENABLE])
            self._send_request(request)

    def set_disable_4pair_power(self, portList):
        self.log_debug('PoePD69100::set_disable_4pair(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        for port in portList:
            status = self._get_single_port_status(port)
            ''' port must be disabled before we can turn off alt-b
            '''
            if status != PD69100PortFullStatusResult.PORT_STATUS_DISABLED:
                self.set_port_disable_poe(port)
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_ALTB_POWER_UP,
                                             port,
                                             PD69100_ALTB_PUP_CFG_ENABLE,
                                             PD69100_ALTB_PUP_CMD_DISABLE])
            self._send_request(request)
            ''' only re-enable if the port wasn't already disabled
            '''
            if status != PD69100PortFullStatusResult.PORT_STATUS_DISABLED:
                self.set_port_enable_poe(portList)

    def set_port_l2_power(self, portList, requested, allocated):
        self.log_debug('PoePD69100::set_port_l2_power(%s, %s, %s)' % (str(portList), requested, allocated))
        portList = self._mk_list(portList)
        requested = int(requested)/100     # .1 watts per tick
        allocated = int(allocated)/100     # .1 watts per tick
        for port in portList:
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                            PD69100_PKT_SUB_CHANNEL,
                                            PD69100_PKT_SUB1_SET_LAYER2_PD, port,
                                            0x53,  # type 2, PSE, PSE
                                            requested >> 8, requested & 0xff,
                                            allocated >> 8, allocated & 0xff,
                                            0,     # 100m (the min)
                                            1,     # execute
                                            ])
            self._send_request(request)

    def get_port_l2_power(self, port):
        request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                        PD69100_PKT_SUB_CHANNEL,
                                        PD69100_PKT_SUB1_GET_LAYER2_PD, port])
        response = self._send_request(request)
        result = {
            'allocated':(response[2] * 256 + response[3])/10,
            'requested':(response[4] * 256 + response[5])/10,
            'pse type':('type1' if response[6] & 0x8 else 'type2'),
            'priority':response[6] & 3,
            'class':response[7] - 1,
            'pair':response[8]
        }
        return result

    def set_device_temp_threshold(self, device, temp):
        request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                        PD69100_PKT_SUB_GLOBAL,
                                        PD69100_PKT_SUB1_DEVICE_PARAMS,
                                        device, PD69100_PKT_DATA_NOP,
                                        temp])
        self._send_request(request)

    def reset_ports(self, portList):
        self.log_debug('PoePD69100::reset_ports(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        reset_ports = []
        for port in portList:
            result = self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PORTSTATUS, port],
                                             PD69100PortFullStatusResult(port),
                                             PD69100PortFullStatusResult.PORT_STATUS_RESPONSE)
            if result.status == result.PORT_STATUS_DISABLED:
                continue

            reset_ports.append(port)
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PORTFULLINIT, port,
                                             PD69100_PKT_DATA_CMDMASKBOTH | PD69100_PKT_DATA_CMDDISABLE,
                                             PD69100_PKT_DATA_AFMASK_AT])
            self._send_request(request)

        time.sleep(5)
        for port in reset_ports:
            request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                             PD69100_PKT_SUB_CHANNEL,
                                             PD69100_PKT_SUB1_PORTFULLINIT, port,
                                             PD69100_PKT_DATA_CMDMASKBOTH | PD69100_PKT_DATA_CMDENABLE,
                                             PD69100_PKT_DATA_AFMASK_AT])
            self._send_request(request)

    ''' pd69100 specific extensions some of these are just for debug purposes
    '''
    def set_power_bank(self, bank, power_limit):
        self.log_debug('PoePD69100::set_power_bank(%u, %u)' % (bank, power_limit))
        request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                         PD69100_PKT_SUB_GLOBAL,
                                         PD69100_PKT_SUB1_SUPPLY,
                                         PD69100_PKT_SUB2_POWER_BUDGET, bank, (power_limit / 256),
                                         power_limit & 0xff,
                                         0x2, 0x3a, 0x01, 0xb8, 0x01])
        self._send_request(request)

    def get_power_bank_settings(self):
        for pb in range(16):
            request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                             PD69100_PKT_SUB_GLOBAL,
                                             PD69100_PKT_SUB1_SUPPLY,
                                             PD69100_PKT_SUB2_POWER_BUDGET, pb])
            self._send_request(request)

    def set_pm_settings(self, pm1, pm2, pm3):
        request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                         PD69100_PKT_SUB_GLOBAL,
                                         PD69100_PKT_SUB1_SUPPLY,
                                         PD69100_PKT_SUB2_POWER_MGMT_MODE,
                                         pm1, pm2, pm3])
        self._send_request(request)

    def set_individual_mask(self, mask, value):
        request = self._new_request_str([PD69100_PKT_KEY_CMD, 0x00,
                                         PD69100_PKT_SUB_GLOBAL,
                                         PD69100_PKT_SUB1_INDIVIDUAL_MASK,
                                         mask, value])
        self._send_request(request)

    def get_individual_masks(self):
        masks = (0x00, 0x01, 0x0c, 0x11, 0x14, 0x15, 0x1b, 0x1e, 0x1f, 0x28, 0x29, 0x2a, 0x2b, 0x2d, 0x2e, 0x2f,
                 0x30, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x43 )
        for mask in masks:
            request = self._new_request_str([PD69100_PKT_KEY_REQUEST, 0x00,
                                             PD69100_PKT_SUB_GLOBAL,
                                             PD69100_PKT_SUB1_INDIVIDUAL_MASK,
                                             mask])
            result = self._send_request(request)

    def get_poe_system_firmware_status(self):
        self.log_debug('PoePD69100::get_poe_system_firmware_status')
        result = PD69100GetSysStatusResult()
        request = self._new_request_str([PD69100_PKT_KEY_REQUEST,
                                         0x00,
                                         PD69100_PKT_SUB_GLOBAL,
                                         PD69100_PKT_SUB1_SYSTEM_STATUS])
        response = self._send_request(request)
        result.handle_response(response, PD69100GetSysStatusResult.CPU_STATUS1_RESPONSE)
        return result.cpu_status1

    def upgrade_firmware(self, fw_file_path):

        # Get system status
        try:
            status = self.get_poe_system_firmware_status()
        except:
            status = 0x0002

        '''Parse system status to determine the state of 
           firmware on the chip. CPU STATUS-1, bit1 = 1
           indicates invalid firmware
        '''
        if status & 0x0002:
            '''Invalid firmware case. Enter program
               mode by sending the following characters.
               Spec recommends keeping at-least 10ms of
               delay b/w them. Kept them at 50ms to be 
               safe.
            '''
            print 'Invalid firmware detected, sending ENTR'
            self.send_ascii_chars_with_wait(["E", "N", "T", "R"], 0.05, PD69100_BOOT_SECTION_RESPONSE)
            print 'Received %s ...[Ok]' % PD69100_BOOT_SECTION_RESPONSE
        else:
            '''Valid firmware detected. Send download
               command as the first step.
            '''
            progkeyresult = PD69100GetProgramKeyResult()
            request = self._new_request_str([PD69100_PKT_KEY_PROG,
                                             0x00,
                                             PD69100_PKT_SUB_FLASH,
                                             0x99, 0x15, 0x16, 0x16, 0x99])
            response = self._send_request(request)
            try:
                if progkeyresult.is_valid_key(response):
                    if progkeyresult.is_checksum_bad(response):
                        raise poe_base.POEDRuntimeError("Checksum failed for download command")
                    elif progkeyresult.is_failed_exec(response):
                        raise poe_base.POEDRuntimeError("Execution failed for download command")
                    else:
                        print "Download command succeeded"
            except poe_base.POEDRuntimeError as e:
                raise e
        try:
            ''' Erase memory
            '''
            time.sleep(1)
            print 'Erasing controller memory'
            self.send_ascii_chars_with_wait(["E"], expected_response = PD69100_ERASE_MEMORY_RESPONSE)
            ''' Wait for erasure. At-least 5 sec.
            '''
            time.sleep(10)
            print 'Received %s ...[Ok]' % PD69100_ERASE_MEMORY_RESPONSE
            ''' Program memory command
            '''
            print 'Sending program memory command'
            self.send_ascii_chars_with_wait(["P"], expected_response = PD69100_PROGRAM_MEMORY_RESPONSE)
            print 'Received %s ...[Ok]' % PD69100_PROGRAM_MEMORY_RESPONSE
            ''' Start sending the firmware .S19 file
            '''
            print "Downloading .S19 file ...",
            sys.stdout.flush()
            with open(fw_file_path, 'r') as fd:
                file_strings_list = fd.readlines()
            for line in file_strings_list:
                if line[:2] != "S0":
                    self.send_ascii_string(line.split('\r\n')[0] + '0D0A')
            print "[Ok]"
            time.sleep(5)
            print "Resetting controller ...",
            sys.stdout.flush()
            self.send_ascii_string(PD69100_RST_REQUEST)
            time.sleep(60)
            print "[Ok]"
        except poe_base.POEDRuntimeError as poe_error:
            raise poe_error
        except IOError as file_error:
            raise file_error

    ''' local utility routines
    '''
    def _get_single_port_status(self, port):
        result = self._handle_port_command([PD69100_PKT_KEY_REQUEST, 0x00, PD69100_PKT_SUB_CHANNEL,
                                         PD69100_PKT_SUB1_PORTSTATUS, port],
                                         PD69100PortFullStatusResult(port),
                                         PD69100PortFullStatusResult.PORT_STATUS_RESPONSE)
        return result.status

    def _persist_settings(self):
        request = self._new_request_str([PD69100_PKT_KEY_PROG, 0x00, PD69100_PKT_SUB_E2,
                                        PD69100_PKT_SUB1_SAVE_CONFIG])
        self._send_request(request)

    def _handle_sys_command(self, cmd_list, result):
        self.log_debug('PoePD69100::_handle_sys_command()')
        request = self._new_request_str(cmd_list)
        response = self._send_request(request)
        result.handle_response(response)
        return result

    def _handle_port_command(self, cmd_list, result, result_type=None):
        self.log_debug('PoePD69100::_handle_port_command()')
        request = self._new_request_str(cmd_list)
        response = self._send_request(request)
        result.handle_response(response, result_type)
        return result

    def _new_request_str(self, cmd_list, no_sequence=False):
        newRequest = []
        for i in range(len(cmd_list)):
            newRequest.append(cmd_list[i])
        for i in range(PD69100_PKT_LEN - PD69100_PKT_CKSUM_LEN - len(newRequest)):
            newRequest.append(0x4e)
        if not no_sequence:
            newRequest[PD69100_PKT_OFFSET_SEQ] = self.seq
            self.seq += 1
            if self.seq == 0xff:
                self.seq = 0
        return newRequest

    def _send_request(self, request):
        retries = 3
        request += self._split_bytes(self._calc_cksum(request))
        while retries:
            try:
                self.lock()
                self._send(request)
                response = self._get_response()
                self.unlock()
                self._verify_response(response, request)
                return response
            except poe_base.POEDRuntimeError, e:
                retries -= 1
        raise e

    def _send(self, request):
        self.log_verbose('tx > %s' % (self._bytes_to_str(request)))
        rlen = len(request)
        if rlen != PD69100_PKT_LEN:
            raise poe_base.POEDRuntimeError('Attempt to send invalid length request - expecting %u, got %u' % (PD69100_PKT_LEN, rlen))
        self.poe_write(request)

    def _get_response(self):
        cstr = self.poe_read(PD69100_PKT_LEN)
        self.log_verbose('rx < %s' % (self._bytes_to_str(cstr)))
        return cstr

    def _clear_rx_queue(self):
        self.poe_read(PD69100_PKT_LEN)

    def _verify_response(self, response, request):
        if len(response) != PD69100_PKT_LEN:
            raise poe_base.POEDRuntimeError('invalid response length')
        elif response[PD69100_PKT_OFFSET_SEQ] != request[PD69100_PKT_OFFSET_SEQ]:
            self._clear_rx_queue()
            raise poe_base.POEDSequenceError('request/response sequence numbers don\'t match [0x%02X:0x%02X]\n' %
                                     (response[PD69100_PKT_OFFSET_SEQ], request[PD69100_PKT_OFFSET_SEQ]))
        calc_cksum = self._calc_cksum(response)
        rx_cksum = response[PD69100_PKT_OFFSET_CKSUM0] * 256 + response[PD69100_PKT_OFFSET_CKSUM1]
        if calc_cksum != rx_cksum:
            raise poe_base.POEDRuntimeError('invalid checksum in response [0x%02X:0x%02X]' % (calc_cksum, rx_cksum))

    def _calc_cksum(self, request):
        rlen = len(request)
        exp_len = PD69100_PKT_LEN - PD69100_PKT_CKSUM_LEN
        if rlen < exp_len:
            raise poe_base.POEDRuntimeError("Calculating checksum on invalid length request - expecting %u, got %u" % (exp_len, rlen))

        sum = 0
        for i in range(exp_len):
            sum += request[i]
        return sum

    def _set_device_reg_bits(poe, reg, val, mask, shift):
        '''
        Set bits in a POE device register. Only use this if you know
        what you are doing! We do a read-modify-write to change only
        the specified bits.

        :param reg: the register to write to
        :param val: the value to write
        :param mask: a mask of the bits to write (shifted)
        :param shift: number of bits to shift the value
        :return: nothing
        '''
        for dev in range(0x0c):
            request = poe._new_request_str([PD69100_PKT_KEY_REQUEST,
                                            0x00,
                                            PD69100_PKT_SUB_TEST,
                                            PD69100_PKT_SUB1_READ_RONS,
                                            reg >> 8,
                                            reg & 0xff,
                                            dev
                                            ])
            response = poe._send_request(request)
            rd = response[2] * 256 + response[3]
            rd = rd & ~mask
            rd |= val << shift
            request = poe._new_request_str([PD69100_PKT_KEY_WRITE,
                                            0x00,
                                            PD69100_PKT_SUB_TEST,
                                            PD69100_PKT_SUB1_WRITE_RONS,
                                            reg >> 8,
                                            reg & 0xff,
                                            rd >> 8,
                                            rd & 0xff,
                                            dev
                                            ])
            poe._send_request(request)

    def _split_bytes(self, bytes):
        return [bytes/256, bytes & 0xff]

    def send_ascii_chars_with_wait(self, char_list, interval = 0.0, expected_response=None):
        for char in char_list:
            self.poe_write_unpacked(char)
            time.sleep(interval)
        if expected_response:
            response_rxd = self.poe_read_unpacked(1500)
            if expected_response not in response_rxd:
                raise poe_base.POEDRuntimeError("Did not receive %s from controller" % expected_response)
        return True

    def verify_ascii_response(self, expected_response):
        response_rxd = self.poe_read_unpacked(1500)
        print response_rxd
        if expected_response not in response_rxd:
            raise poe_base.POEDRuntimeError("Did not receive %s from controller" % expected_response)
        return True

    def send_ascii_string(self, string, expected_response=None):
        self.poe_write_unpacked(string)
        if expected_response:
            response_rxd = self.poe_read_unpacked(1500)
            if expected_response not in response_rxd:
                raise poe_base.POEDRuntimeError("Did not receive %s from controller" % expected_response)
        return True


class PD69100PortStatusResult(poe_base.PortStatusResult):

    def handle_response(self, response):
        if self.poe_port < 22:
            port_byte = response[PD69100_PKT_OFFSET_RXDATA0 + self.poe_port % 11]
        elif self.poe_port < 24:
            port_byte = response[PD69100_PKT_OFFSET_RXDATA0 + (self.poe_port - 22)]
        elif self.poe_port < 26:
            port_byte = response[PD69100_PKT_OFFSET_RXDATA0 + (self.poe_port - 19)]
        else:
            port_byte = response[PD69100_PKT_OFFSET_RXDATA0 + (self.poe_port - 26) % 11]
        self.status = pd69100_port_ok_results.get(port_byte, poe_base.PortStatusResult.PORT_STATUS_FAULT)
        if self.status == poe_base.PortStatusResult.PORT_STATUS_FAULT:
            self._error_str = pd69100_port_error_map.get(port_byte, 'other [0x%02X]' % (port_byte))
        elif self.status == poe_base.PortFullStatusResult.PORT_STATUS_DELIVERING_POWER:
            self.pd_type = pd69100_port_pd_type_map.get(port_byte, port_byte)

class PD69100PortFullStatusResult(poe_base.PortFullStatusResult):
    MEASUREMENTS_RESPONSE=1
    PORT_STATUS_RESPONSE=2
    PRIORITY_RESPONSE=3

    def handle_response(self, response, response_type):
        if response_type == self.PORT_STATUS_RESPONSE:
            en_dis = response[PD69100_PKT_OFFSET_RXDATA0]
            if en_dis == PD69100_PKT_DATA_PORT_DISABLED:
                self.status = poe_base.PortFullStatusResult.PORT_STATUS_DISABLED
            else:
                port_byte = response[PD69100_PKT_OFFSET_RXDATA1]
                self.status = pd69100_port_ok_results.get(port_byte, poe_base.PortFullStatusResult.PORT_STATUS_FAULT)
                if self.status == poe_base.PortFullStatusResult.PORT_STATUS_FAULT:
                    self._error_str = pd69100_port_error_map.get(port_byte, 'other [0x%02X]' % (port_byte))
                elif self.status == poe_base.PortFullStatusResult.PORT_STATUS_DELIVERING_POWER:
                    self.pd_class = response[PD69100_PKT_OFFSET_RXDATA4]
                    self.pd_type = pd69100_port_pd_type_map.get(port_byte, port_byte)
        elif response_type == self.MEASUREMENTS_RESPONSE:
            self.current = self._mk_word_data(response[PD69100_PKT_OFFSET_RXDATA2], response[PD69100_PKT_OFFSET_RXDATA3])           # mA
            self.power = self._mk_word_data(response[PD69100_PKT_OFFSET_RXDATA4], response[PD69100_PKT_OFFSET_RXDATA5])/1000.0      # Watts
            val = self._mk_word_data(response[PD69100_PKT_OFFSET_RXDATA7], response[PD69100_PKT_OFFSET_RXDATA8])
            if val == 65535:
                val = 0
            self.voltage = val/10.0      # Volts

        elif response_type == self.PRIORITY_RESPONSE:
            self.priority = pd69100_priority_map.get(response[PD69100_PKT_OFFSET_RXDATA0])

class PD69100PortPriorityResult(poe_base.PortPriorityResult):
        def handle_response(self, response, type):
            self.priority = pd69100_priority_map.get(response[PD69100_PKT_OFFSET_RXDATA0])

class PD69100GetSysPowerResult(poe_base.PoeResult):

    def handle_response(self, response):
        self.sys_allocated_power = self._mk_word_data(response[PD69100_PKT_OFFSET_RXDATA0], response[PD69100_PKT_OFFSET_RXDATA1])
        self.sys_available_power = self._mk_word_data(response[PD69100_PKT_OFFSET_RXDATA8], response[PD69100_PKT_OFFSET_RXDATA9])

class PD69100GetSysStatusResult(poe_base.PoeSysStatusResult):
    INTERRUPTS_RESPONSE=1
    SUPPLY_RESPONSE=2
    POWER_RESPONSE=3
    TEMP_RESPONSE=4
    CPU_STATUS1_RESPONSE=5

    def handle_response(self, response, response_type):
        if response_type == self.INTERRUPTS_RESPONSE:
            int_val = response[PD69100_PKT_OFFSET_RXDATA9] * 256 + response[PD69100_PKT_OFFSET_RXDATA10]
            if int_val & PD69100_INT_SYS_EVENT_MASK:
                if int_val & PD69100_INT_DEVICE_OVERTEMP1 or int_val & PD69100_INT_DEVICE_OVERTEMP2:
                    self._error_str = 'System Overtemperature'
                elif int_val & PD69100_INT_INTERNAL_DEVICE_FAULT:
                    self._error_str = 'Internal System Fault'
                elif int_val & PD69100_INT_MAIN_VOLTAGE_ERROR:
                    self._error_str = 'System Main Voltage Error'
                else:
                    self._error_str = 'Unknown System Error [0x%04X]' % (int_val)
            if int_val & PD69100_INT_PORT_EVENT_MASK:
                self._port_event = True
        elif response_type == self.SUPPLY_RESPONSE:
            self.sys_consumed_power = self._mk_word_data(response[PD69100_PKT_OFFSET_RXDATA0], response[PD69100_PKT_OFFSET_RXDATA1])
            self.sys_power_limit = self._mk_word_data(response[PD69100_PKT_OFFSET_RXDATA8], response[PD69100_PKT_OFFSET_RXDATA9])
        elif response_type == self.TEMP_RESPONSE:
            temp = response[PD69100_PKT_OFFSET_RXDATA7]
            if temp != 255:
                self.sensors[response[PD69100_PKT_OFFSET_RXDATA0]] = temp * 1000
        elif response_type == self.CPU_STATUS1_RESPONSE:
            self.cpu_status1 = response[PD69100_PKT_OFFSET_RXDATA0]

class PD69100GetPort4pairStateResult(poe_base.PoePort4portStateResult):
    def handle_response(self, response, type=None):
        if response[PD69100_PKT_OFFSET_RXDATA1] != 0:
            self.pair_mode = self.RESULT_4_PAIR
        else:
            self.pair_mode = self.RESULT_2_PAIR

class PD69100GetSwVersionResult(poe_base.SwVersionResult):

    def handle_response(self, response):
        self.product_num = response[PD69100_PKT_OFFSET_RXDATA2]
        sw_ver = response[PD69100_PKT_OFFSET_RXDATA3] * 256 + response[PD69100_PKT_OFFSET_RXDATA4]
        sw_ver_str = '%04u' % (sw_ver)
        self.major_version = int(sw_ver_str[:2])
        self.minor_version = int(sw_ver_str[2])
        self.patch = int(sw_ver_str[3])
        self.ver_str = '%02X.%02u.%u.%u' % (self.product_num, self.major_version, self.minor_version, self.patch)

class PD69100GetDiagInfoResult(poe_base.DiagInfoResult):
    SUPPLY_RESPONSE=1
    MODE_RESPONSE=2
    MASK_RESPONSE=3
    REG_RESPONSE=4

    def __init__(self):
        poe_base.DiagInfoResult.__init__(self)
        self.individual_masks = {}
        self.ports = {}
        self.startup_to = {}

    def handle_response(self, response, response_type, response_info=None):
        if response_type == self.SUPPLY_RESPONSE:
            self.supply_power_consumption = response[PD69100_PKT_OFFSET_RXDATA0] * 256 + response[PD69100_PKT_OFFSET_RXDATA1]
            self.supply_max_shutdown_voltage = response[PD69100_PKT_OFFSET_RXDATA2] * 256 + response[PD69100_PKT_OFFSET_RXDATA3]
            self.supply_min_shutdown_voltage = response[PD69100_PKT_OFFSET_RXDATA4] * 256 + response[PD69100_PKT_OFFSET_RXDATA5]
            self.supply_power_bank = response[PD69100_PKT_OFFSET_RXDATA7]
            self.supply_power_limit = response[PD69100_PKT_OFFSET_RXDATA8] * 256 + response[PD69100_PKT_OFFSET_RXDATA9]
        elif response_type == self.MODE_RESPONSE:
            self.mode_pm1 = response[PD69100_PKT_OFFSET_RXDATA0]
            self.mode_pm2 = response[PD69100_PKT_OFFSET_RXDATA1]
            self.mode_pm3 = response[PD69100_PKT_OFFSET_RXDATA2]
        elif response_type == self.MASK_RESPONSE:
            self.individual_masks['0x%02X' % response_info] = '%u' % (response[PD69100_PKT_OFFSET_RXDATA0])
        elif response_type == self.REG_RESPONSE:
            val = response[PD69100_PKT_OFFSET_RXDATA0]*256 + response[PD69100_PKT_OFFSET_RXDATA1]
            self.startup_to[response_info] = ((val & 0xfe0) >> 5)*4        # timeout in ms

class PD69100GetPortDiagInfoResult(poe_base.PoePortResult):
    STATUS_RESPONSE=1
    POWER_LIMIT_RESPONSE=2
    L2_POWER_RESPONSE=3

    def handle_response(self, response, response_type):
        if response_type == self.STATUS_RESPONSE:
            self.status_enabled = '0x%02X' % response[PD69100_PKT_OFFSET_RXDATA0]
            self.status_status = '0x%02X' % response[PD69100_PKT_OFFSET_RXDATA1]
            self.status_force_power = '0x%02X' % response[PD69100_PKT_OFFSET_RXDATA2]
            self.status_latch = '0x%02X' % response[PD69100_PKT_OFFSET_RXDATA3]
            self.status_class = '0x%02X' % response[PD69100_PKT_OFFSET_RXDATA4]
            self.status_af_at = '0x%02X' % response[PD69100_PKT_OFFSET_RXDATA8]
            self.status_4pair = '0x%02X' % response[PD69100_PKT_OFFSET_RXDATA9]
        elif response_type == self.POWER_LIMIT_RESPONSE:
            self.power_ppl = '0x%04X' % (response[PD69100_PKT_OFFSET_RXDATA0] * 256 + response[PD69100_PKT_OFFSET_RXDATA1])
            self.power_tppl = '0x%04X' % (response[PD69100_PKT_OFFSET_RXDATA2] * 256 + response[PD69100_PKT_OFFSET_RXDATA3])
        elif response_type == self.L2_POWER_RESPONSE:
            self.l2_allocated = '%.1fw' % ((response[PD69100_PKT_OFFSET_RXDATA0] * 256 + response[PD69100_PKT_OFFSET_RXDATA1])/10)
            self.l2_requested = '%.1fw' % ((response[PD69100_PKT_OFFSET_RXDATA2] * 256 + response[PD69100_PKT_OFFSET_RXDATA3])/10)
            self.l2_actual = '%.1fw' % (((response[PD69100_PKT_OFFSET_RXDATA9] * 256 + response[PD69100_PKT_OFFSET_RXDATA10]) & 0xfff)/10)

class PD69100GetProgramKeyResult(poe_base.ProgramKeyResult):

    def handle_response(self, response):
        key = response[PD69100_PKT_OFFSET_KEY]
        sub = response[PD69100_PKT_OFFSET_RXDATA0]
        sub1 = response[PD69100_PKT_OFFSET_RXDATA1]
        sub2 = response[PD69100_PKT_OFFSET_RXDATA2]
        data = response[PD69100_PKT_OFFSET_RXDATA3]
        if key == PD69100_PKT_KEY_REPORT:
            success = (sub == 0x00 and sub1 == 0x00)
            cksum_failed = (sub == 0xff and sub1 == 0xff and sub2 == 0xff and data == 0xff)
            exec_failed = False
            if not success and not cksum_failed:
                exec_failed = True
            self.result['PASS'] = success
            self.result['BAD_CKSUM'] = cksum_failed
            self.result['EXEC_FAIL'] = exec_failed
        else:
            raise poe_base.POEDRuntimeError("Invalid key type(%x) detected in response to program key" %key)

    def is_checksum_bad(self, response):
        try:
            self.handle_response(response)
        except poe_base.POEDRuntimeError as e:
            raise e
        return self.result['BAD_CKSUM']

    def is_successful(self, response):
        try:
            self.handle_response(response)
        except poe_base.POEDRuntimeError as e:
            raise e
        return self.result['PASS']

    def is_failed_exec(self, response):
        try:
            self.handle_response(response)
        except poe_base.POEDRuntimeError as e:
            raise e
        return self.result['EXEC_FAIL']

    def is_valid_key(self, response):
        try:
            self.handle_response(response)
        except poe_base.POEDRuntimeError as e:
            raise e
        return True
